"""
    Core
"""

# Core
from .core import ID  # Scalar
from .core import Controller  # Class
from .core import Date  # Class
from .core import date  # Scalar
from .core import datetime  # Scalar
from .core import decimal  # Scalar
from .core import field  # Tool - Field
from .core import is_form  # Tool | is-form?
from .core import is_model  # Tool | is-model?
from .core import json  # Scalar
from .core import load  # Tool - Load
from .core import text  # Scalar
from .core import time  # Scalar
from .core import type  # Type

# Forms
from .forms import Form as form  # Tool - Create Forms
from .manager import manager  # Tool - Querying Manager



"""Names & Info"""


def get_names(part):
    """Get Operations Names"""
    return list(map(lambda x: x["name"], part["fields"]))


def get_info(schema):
    """Get GraphQL (Query & Mutations)"""
    app = schema["__schema"]
    return {
        "query": get_names(app["query"]),
        "mutation": get_names(app["mutation"]),
    }



"""
    On Events (FastAPI)
"""
from fastapi.responses import RedirectResponse


def redirect_root(context):
    """Redirect"""

    @context.app.get("/", response_class=RedirectResponse)
    async def redirect_to_docs():
        """Redirect To Docs"""
        return "/docs"

    return context



"""empty message

Revision ID: 71485f40a569
Revises: 
Create Date: 2022-11-27 22:52:42.848655

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '71485f40a569'
down_revision = None
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('todo_task',
    sa.Column('_id', sa.Integer(), nullable=False),
    sa.Column('title', sa.String(length=255), nullable=True),
    sa.Column('description', sa.String(length=255), nullable=True),
    sa.Column('status', sa.String(length=255), nullable=True),
    sa.PrimaryKeyConstraint('_id')
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('todo_task')
    # ### end Alembic commands ###



"""
    Mongo Manager
"""
import functools
import typing
from collections import OrderedDict, namedtuple

from .. import manager

# from ..core.custom_class import Globals

try:
    import motor.motor_asyncio

    SQLALCHEMY_ACTIVE = True
    PM = motor.motor_asyncio
except ImportError:
    SQLALCHEMY_ACTIVE = False
    PM = None


MongoDNS = namedtuple("MongoDNS", ["url", "database"])
DatabaseManager = namedtuple("Mongo", ["base", "database", "controller"])


def mongo_fields(annotations, config) -> OrderedDict:
    """All Column(s) Setup"""
    db_fields = OrderedDict({"_id": True})
    related_fields = OrderedDict()
    for field in annotations:
        if field.is_custom_type:
            if field.name not in config["ignore"]:
                related_fields[f"{field.name}_id"] = True
        else:
            if field.name not in config["ignore"]:
                db_fields[field.name] = True

    for key in related_fields.keys():
        db_fields[key] = True
    return db_fields


def mongo_model(
    controller: typing.Any = None,
    table_name: str = None,
    annotations: typing.Any = None,
    config: dict = None,
    new_object: typing.Any = None,
):
    """Mongo ORM Setup"""
    fields = mongo_fields(annotations, config)
    objects = controller.base[table_name]
    # Finally
    db_keys = list(fields.keys())
    new_object.__database__ = "mongo"
    new_object.objects = controller.controller(objects, keys=db_keys)
    return new_object


def get_database_name(url):
    """Get Table's Name from URL"""
    items = url.split("/")[-1::]
    return items[0] if len(items) == 1 else None


def get_database_url(url):
    """Get Database's URL"""
    return "/".join(url.split("/")[:-1])


def get_database_info(url):
    """{ get_database_name } + { get_database_url }"""
    return MongoDNS(url=get_database_url(url), database=get_database_name(url))


def create_mongo_manager(url: str):
    """Create Mongo Base-Manager"""
    info = get_database_info(url)
    engine = PM.AsyncIOMotorClient(info.url)
    database = engine[info.database]
    controller = functools.partial(manager.Mongo, database=database)
    return DatabaseManager(base=database, database=database, controller=controller)



"""
    Custom Class Creator
"""
import dataclasses as dc
import typing
from collections import OrderedDict

from .scalars import SCALAR_FIELD, STRAWBERRY_CORE


class Globals:
    """API Globals"""

    @staticmethod
    def set(key, value):
        """Set"""
        globals().update({key: value})

    @staticmethod
    def get(key):
        """Get"""
        return globals().get(key)


class BaseClass:
    """API Base Class"""

    _id: typing.Optional[str] = None
    id: typing.Optional[SCALAR_FIELD.ID] = None


def create_class(custom_class, original_object, model):
    """Create Class for API"""
    ordered_annotations = OrderedDict()
    unordered_annotations = {
        **BaseClass.__annotations__,
    }
    dataclass = dc.dataclass(original_object)
    fields = dc.fields(dataclass)
    for field in fields:
        if field.name not in model.methods:
            f_default = isinstance(field.default, dc._MISSING_TYPE)
            f_default_factory = isinstance(field.default_factory, dc._MISSING_TYPE)
            out_default = dc.field(default=None)
            out_type = model.annotations[field.name].graphql
            if not f_default and f_default_factory:
                out_default = dc.field(default=field.default)
            elif f_default and not f_default_factory:
                out_default = dc.field(default_factory=field.default_factory)
            setattr(custom_class, field.name, out_default)
            unordered_annotations[field.name] = out_type

    # Set Method Fields
    for field in model.methods:
        current_method = getattr(original_object, field)
        if STRAWBERRY_CORE:
            current_method = STRAWBERRY_CORE.field(current_method)
            setattr(custom_class, field, current_method)

    # Create Annotations
    for field in ["_id", "id"] + model.fields:
        anno = unordered_annotations[field]
        ordered_annotations[field] = anno

    custom_class.__annotations__ = {**ordered_annotations}
    return custom_class


def create_custom_type(original_object, model, description=None):
    """Create Custom Class"""
    class_name = original_object.__name__
    custom_class = type(
        class_name,
        (BaseClass,),
        {},
    )
    custom_class = create_class(custom_class, original_object, model)

    # Strawberry
    if STRAWBERRY_CORE:
        description = description or original_object.__doc__
        custom_class = STRAWBERRY_CORE.type(custom_class, description=description)
    elif not STRAWBERRY_CORE:
        custom_class = dc.dataclass(custom_class)

    # Register Global
    Globals.set(class_name, custom_class)

    return custom_class



"""
    Scalars
"""
import dataclasses as dc
import datetime
import decimal
import types
import typing

# Strawberry
try:
    import strawberry

except ImportError:
    strawberry = None


# Custom Typing
TEXT = typing.TypeVar("TEXT", str, None)
JSON = typing.TypeVar("JSON", object, dict, list, None)
ID = typing.TypeVar("ID", str, int, None)


# Map GQL
STRAWBERRY_CORE = False
SCALAR_FIELD = types.SimpleNamespace(**{key: key for key in ["ID", "TEXT", "JSON"]})
SCALAR_FIELD.TEXT = str
SCALAR_FIELD.JSON = dict
SCALAR_FIELD.ID = str
if strawberry:
    STRAWBERRY_CORE = strawberry
    SCALAR_FIELD.JSON = strawberry.scalars.JSON
    SCALAR_FIELD.ID = strawberry.ID

# Global
SCALAR_FIELD_ID = SCALAR_FIELD.ID


@dc.dataclass(frozen=True)
class APIType:
    """API-Field-Type"""

    name: str
    python: typing.Any
    # real: typing.Any


# Core Init
Scalar = {}

# Core Values
Scalar[ID] = APIType(
    name="ID",
    python=SCALAR_FIELD.ID,
)

Scalar[TEXT] = APIType(
    name="Text",
    python=SCALAR_FIELD.TEXT,
)

Scalar[JSON] = APIType(
    name="JSON",
    python=SCALAR_FIELD.JSON,
)

Scalar[str] = APIType(
    name="String",
    python=str,
)

Scalar[int] = APIType(
    name="Integer",
    python=int,
)

Scalar[float] = APIType(
    name="Float",
    python=float,
)

Scalar[bool] = APIType(
    name="Boolean",
    python=bool,
)

Scalar[datetime.date] = APIType(
    name="Date",
    python=datetime.date,
)

Scalar[datetime.datetime] = APIType(
    name="DateTime",
    python=datetime.datetime,
)

Scalar[datetime.time] = APIType(
    name="Time",
    python=datetime.time,
)

Scalar[decimal.Decimal] = APIType(
    name="Decimal",
    python=decimal.Decimal,
)



"""
    * Response
"""

from dataclasses import dataclass
from typing import Any


@dataclass
class Response:
    """[summary]
    Output from the database (SQL or MongoDB) response.
    """

    data: Any = None
    error: bool = False
    error_message: str = None
    count: int = 0
    pages: int = 0



# -*- coding: utf-8 -*-
"""
    Custom - Command-Line-Group
"""
import os
from pathlib import Path

import fastberry as fb

from .utils import Frontend, write_file

JS_README = """
# FastBerry (ESM)

## Init the **API**

```js
import client from "./client";

const api = client({
  url: "http://localhost:8000/graphql",
  debug: true,
  maxDepth: 6,
  config: {
    mode: "cors",
    credentials: "same-origin", // same-origin include
    headers: {
      Accept: "application/json",
      "Content-Type": "application/json; charset=UTF-8",
    },
  },
});

// Pre-Built { Query(s) & Mutation(s) } from { Schema }
console.log(api.admin.keys);
```

---

## How To **Disable Fields**

```js
// Single Model
const instance = api.gql("query", "MyQuery", ["Id"]);

// Edge Model
const instance = api.gql("query", "MyQuery", ["edges.node.Id"]); // Recursive Disable ["via.dots"]

// Multi Model
const instance = api.gql("mutation", "MyMutation", { Model: ["Id"] });
```

---

## **GQL** Request Demo

```js
// Init (Editable) Instance
const instance = api.gql("query", "MyQuery", ["Id"]);

// Edit
instance.form.item = "MTo6YTU1ZTUzMmVhYjAyOGI0Mg==";

// Run
const { MyQuery } = await instance.run(instance);

console.log(MyQuery);
```

---

## **Chain** Request Demo

```js
const request = {
  MyQuery: [
    // Response
    {
      id: true,
      name: true,
    },
    // Inputs
    {
      item: "MTo6YTU1ZTUzMmVhYjAyOGI0Mg==",
    },
  ],
};

api.chain("query", request).then((response) => {
  console.log(response);
});
```
"""

# Create <Commands> here.
def build_client(schema):
    """Demo CLI Function"""

    # "Generates" Folder
    root_folder = fb.config["spoc"]["spoc"]["generates"]

    # GQL Folder
    root_folder = Path(root_folder)
    folder = root_folder / "client"

    # Schema
    frontend = Frontend(schema)

    # Main JS
    main_js = """import models from "./models";\n\n"""
    main_js += f"const schema = {frontend.schema}\n\n"
    main_js += """
    function getObj(t,e,n,r=0){const s={};return Object.keys(n).forEach((o=>{const i=n[o];if(!(r>e))if(!0===i)s[o]=i;else if(!1===i);else{const n=t.model[i];s[o]=getObj(t,e,n,r+1)}})),s}function getModel(t,e,n=2){return getObj(t,n,e)}function getForms(t,e,n,r){const s={};return r.forEach((r=>{const o=t[e][r][n],i="{}"===JSON.stringify(o)||"[]"===JSON.stringify(o);s[r]=i?null:Object.freeze(o)})),Object.freeze(s)}function getModels(t,e){const n={};return e.forEach((e=>{const r=t[e].model;n[e]=r||!0})),Object.freeze(n)}function createApiChain(t,e,n,r){const s={};return s[n]=[r.model,r.form],t.chain(e,s)}function setValueByPath(t,e,n){let r=t;return function t(e,n,r){let s=n.shift();e.hasOwnProperty(s)&&(0===n.length?e[s]=r:t(e[s],n,r))}(r,e.split("."),n),r}function getGraphqlSetup(t,e,n,r){const s={...t.form[e][n]},o=t.returnType[e][n];let i=null;if(o)if(1===o.length){const e=o[0];let n=t.model(e);n=JSON.parse(JSON.stringify(n)),r&&r.forEach((t=>{n=setValueByPath(n,t,!1)})),i=n}else o.length>1&&(i={},o.forEach((e=>{const n=`...on ${e}`;let s=t.model(e);s=JSON.parse(JSON.stringify(s)),r&&r[e]&&r[e].forEach((t=>{s=setValueByPath(s,t,!1)})),i[n]=s})));return{form:s,model:i,run:r=>createApiChain(t.chain,e,n,r)}}class ClientGQL{constructor(t,e=4,n){this.chain=n,this._schema=Object.freeze(t),this._maxDepth=e,this._keys={query:Object.keys(t.query),mutation:Object.keys(t.mutation),model:Object.keys(t.model)},this._forms={query:getForms(t,"query","form",this.keys.query),mutation:getForms(t,"mutation","form",this.keys.mutation)},this._args={query:getForms(t,"query","args",this.keys.query),mutation:getForms(t,"mutation","args",this.keys.mutation)},this._returnType={query:getModels(t.query,this.keys.query),mutation:getModels(t.mutation,this.keys.mutation)}}get dir(){return["schema","keys","create","form","args","returnType"]}get schema(){return this._schema}get keys(){return this._keys}get form(){return this._forms}get args(){return this._args}get returnType(){return this._returnType}create(t,e,n=null){return getGraphqlSetup(this,t,e,n)}model(t,e=null){const n=e||this._maxDepth;return getModel(this.schema,this.schema.model[t],n)}}function createGraphQLRequest(t,e){function n(t){return JSON.stringify(function(t){return"number"==typeof t?t:parseFloat(t)||parseInt(t)||t}(t))}function r(t,e,r){let s=`${function(t){if("object"==typeof t){const e=Object.keys(t)[0];return e+" : "+t[e]}return t}(t)}`;return s+=function(t){let e="";if(t){e+="(";const r=[];Object.keys(t).forEach((e=>{const s=[];let o=t[e];o instanceof Object?(Object.keys(o).forEach((t=>{s.push(`${t}: ${n(o[t])}`)})),r.push(`${e}: { ${s} }`)):r.push(`${e}: ${n(o)}`)})),e+=r.join(", "),e+=")"}return e}(r),s+=" "+function(t){let e="{";return function t(n,r){Object.keys(n).forEach((r=>{"object"==typeof n[r]?(e+=` ${r} {`,t(n[r],r),e+=" }"):n[r]&&(e+=` ${r}`)}))}(t),e+=" }",e}(e),s}let s=`${t} { `;return Object.keys(e).forEach((t=>{const n=e[t],o=r(t,n[0],n[1]);s+=" ",s+=o})),s+=" }",s.trim()}function APIHandler(t,e){return new Promise(((n,r)=>{fetch(t,e).then((t=>t.json())).then((t=>n(t))).catch((t=>r(t)))}))}class API{constructor(t,e,n){this._debug=null!==n&&n,this._host=t,this._options=e}post(t=!1){const e={method:"POST"};t&&(e.body=JSON.stringify(t));const n={...this._options,...e};return APIHandler(this._host,n)}graphql(t=null){return this.post({query:t})}async chain(t,e){const n=createGraphQLRequest(t,e),{data:r,errors:s}=await this.graphql(n);return this._debug&&s&&console.error(s),r}}function createAPI(t,e={},n=null){return new API(t,{mode:"cors",credentials:"same-origin",headers:{Accept:"application/json","Content-Type":"application/json; charset=UTF-8"},...e},n)}function GraphQL(t,e,n,r){return new ClientGQL({...e,model:n},r,t)}function isBlank(t,e=null){return t||e}function Api(t){let e=null,n=null;const r=null!=t.debug&&t.debug,s=isBlank(t.maxDepth,4);return e=createAPI(t.url,t.config,r),t.schema&&t.models&&(n=GraphQL(e,t.schema,t.models,s)),{admin:n,chain:(...t)=>e.chain(...t),gql:(...t)=>n.create(...t)}}export default function Wrap(t){return"object"==typeof t?Api({schema:schema,models:models,...t}):createAPI(t)}
    """.strip()

    # Schema
    # write_file(root_folder / "schema.graphql", str(schema))

    # README
    write_file(root_folder / "README.md", JS_README)

    # INDEX
    write_file(folder / "index.js", main_js)

    # Models
    write_file(folder / "models.js", f"export default {frontend.models}")

    # Run Build
    os.chdir(folder)



"""
    GraphQL (Schema + Operations)
"""

import datetime
import decimal
import functools
import json
import os
import shutil

import click

from ...tools import to_camel_case
from .gql import build_client, introspection_info
from .shell import shell_print, unzip, zip


def get_controller():
    """Get Controller"""
    from ... import App

    return App()


SCALAR_INPUT = {
    list: "Array",
    str: "String",
    int: "Integer",
    float: "Float",
    bool: "Boolean",
    decimal.Decimal: "Decimal",
    datetime.date: "Date",
    datetime.datetime: "Datetime",
    datetime.time: "Time",
}


def reset_folder(folder):
    """Reset Folder"""
    for filename in os.listdir(folder):
        file_path = os.path.join(folder, filename)
        try:
            if os.path.isfile(file_path) or os.path.islink(file_path):
                os.unlink(file_path)
            elif os.path.isdir(file_path):
                shutil.rmtree(file_path)
        except Exception as e:
            print(f"Failed to delete {file_path}. Reason: {e}")


def create_graphql_types(types):
    """Get All Custom-Types"""

    setup = {}
    for model in types:
        class_name = model.__name__
        fields = model.__spoc__.config["annotations"]
        setup[class_name] = {"id": True}
        for key, val in fields.items():
            camel_key = to_camel_case(key)
            if val.scalar:
                setup[class_name][camel_key] = True
            else:
                try:
                    forward_name = val.real.__forward_arg__
                except:
                    forward_name = val.real
                setup[class_name][camel_key] = forward_name
    return setup


def form_default_value(value, input_type):
    """Form: Get Default Value"""
    output = None
    if input_type == "string":
        output = ""
    elif value == list:
        output = []
    return output


def create_graphql_forms(forms):
    """Get All Custom-Forms"""

    setup = {}
    for cls in forms:
        annotations = cls.__spoc__.config["annotations"]
        name = cls.__name__
        setup[name] = {
            "name": name,
            "fields": [],
            "labels": {},
            "form": {},
        }
        for key, val in annotations.items():
            camel_key = to_camel_case(key)
            input_type = SCALAR_INPUT.get(val)
            input_type = input_type or "string"
            value = form_default_value(val, input_type)
            field = {
                "name": camel_key,
                "type": input_type,
            }
            setup[name]["fields"].append(field)
            setup[name]["labels"][camel_key] = key.title().replace("_", " ")
            setup[name]["form"][camel_key] = value

    return setup


def pretty(items):
    """Pretty (Operations)"""
    return sorted(list(set(items)))


def clean_print_name(base_dir, active):
    """Clean Print Name"""
    return os.path.relpath(active, base_dir)


def write_json(path, data, javascript=False):
    """Write JSON File"""
    with open(path, "w", encoding="utf-8") as file:
        json_data = json.dumps(data, indent=4, sort_keys=True)
        if javascript:
            json_data = f"export default {json_data}"
        file.write(json_data)


def write_schema(path, data):
    """Write Schema File"""
    with open(path, "w", encoding="utf-8") as file:
        file.write(str(data))


def go_to_path(base_dir, user_dir, default_dir="graphql"):
    """Go To Path By String"""
    out_dir = base_dir
    if user_dir == "" or not user_dir:
        user_dir = default_dir
    for path in user_dir.split("/"):
        out_dir = out_dir / path
    return out_dir


@click.command(name="schema")
@click.option(
    "-c",
    "--client",
    default=False,
    is_flag=True,
    show_default=True,
    help="Create JavaScript GraphQL.",
)
def graphql(client):
    """Build GraphQL (Schema + Operations)"""
    controller = get_controller()

    # Settings Path(s)
    generates_dir = controller.core.config["spoc"].get("spoc", {}).get("generates")

    # Get Path(s)
    base_dir = controller.core.base_dir
    output_dir = go_to_path(base_dir, generates_dir, "graphql")

    # Create Path(s)
    output_dir.mkdir(parents=True, exist_ok=True)

    # Clean
    reset_folder(output_dir)

    # Output Paths <Files>
    codeout = {
        "schema": output_dir / "schema.graphql",
        "schema-ops": output_dir / "operations.json",
    }

    clean_name = functools.partial(clean_print_name, output_dir)

    # Core Information
    messages = {
        "main-folder": "[Building]: Schema & Operations . . .",
        "folder": f"[Folder]: {output_dir}",
    }

    size = len(messages["folder"])
    messages["cool-folder"] = "-" * size

    # GraphQL (Schema)
    schema = controller.graphql.schema()
    schema_info = introspection_info(schema)

    # Names (Operations)
    # query_names = controller.graphql.info.query
    # mutation_names = controller.graphql.info.mutation
    # query_names_camel = set()
    # mutation_names_camel = set()

    # for x in query_names:
    #     query_names_camel.add(to_camel_case(x))
    # for x in mutation_names:
    #     mutation_names_camel.add(to_camel_case(x))

    # operations_names = {
    #     "query": pretty(query_names_camel),
    #     "mutation": pretty(mutation_names_camel),
    # }

    if client:
        # Fastberry GraphQL
        shell_print(messages["cool-folder"])
        shell_print("[FastBerry]: Creating GraphQL Client...")
        shell_print(messages["cool-folder"])
        shell_print(messages["folder"])
        shell_print(messages["cool-folder"])
        shell_print(
            """* [Folder] : client/""",
            color="yellow",
        )
        shell_print(
            """* [ReadMe] : README.md """,
            color="yellow",
        )
        build_client(schema)
    else:
        # Core GraphQL
        shell_print(messages["cool-folder"])
        shell_print(messages["main-folder"])
        shell_print(messages["cool-folder"])
        shell_print(messages["folder"])
        shell_print(messages["cool-folder"])
        shell_print(
            f"""* [Schema]         : { clean_name(codeout["schema"]) }""",
            color="yellow",
        )
        shell_print(
            f"""* [Operations]     : { clean_name(codeout["schema-ops"]) }""",
            color="yellow",
        )
        # Outputs
        write_schema(codeout["schema"], schema)
        write_json(codeout["schema-ops"], schema_info)



"""
    Tool to Transform { Text } to { Code-Cases }
    Tool to Get Module's Name
"""
import re
import typing


def to_kebab_case(text):
    """Clean Text"""
    text = re.sub("[^0-9a-zA-Z]+", "-", str(text))
    return text.lower()


def to_camel_case(text):
    """Converts to { camelCase }"""
    text = to_kebab_case(text)
    init, *temp = text.title().split("-")
    return "".join([init.lower()] + temp)


def to_pascal_case(text):
    """Converts to { PascalCase }"""
    text = to_kebab_case(text)
    return text.title().replace("-", "")


def get_module_name(model: typing.Any):
    """Get: Class-Module's Name"""
    name = model.__module__
    parts = name.split(".")
    if parts[0] == "apps" and len(parts) > 1:
        parts.pop(0)
    module_name = parts[0]
    if module_name == "__main__":
        module_name = None
    return module_name



"""
    Framework Core Managers To Transform Blocks of Code Into A Full-Project
"""
from .api_info import api_info
from .commands import commands
from .extras import extras
from .forms import forms
from .graphql import graphql
from .pagination import pagination
from .routers import APIRouter, routers
from .types import types



"""
    [GraphQL-Tools]
"""

from .client import build_client
from .reader import get_info

INTROSPECTION_QUERY = """
{
  __schema {
    mutation: mutationType {
      name
      fields {
        name
      }
    }
    query: queryType {
      name
      fields {
        name
      }
    }
  }
}
"""


def introspection_info(schema):
    """GraphQL Introspection INFO"""
    result = schema.execute_sync(INTROSPECTION_QUERY)
    return get_info(result.data)



"""
    Core
"""

# App
from .app import DATE as date
from .app import DATETIME as datetime
from .app import DECIMAL as decimal
from .app import TIME as time
from .app import TYPE as type
from .app import Controller, Date, field, load

# Scalars
from .scalars import JSON as json
from .scalars import SCALAR_FIELD_ID as ID
from .scalars import TEXT as text
from .spoc import is_form, is_model



"""
    App
"""
import dataclasses as dc
import datetime
import decimal
import functools

from ..models import create_database
from ..tools import get_module_name
from .base import custom_type
from .spoc import is_model

DATE = datetime.date
DATETIME = datetime.datetime
TIME = datetime.time
DECIMAL = decimal.Decimal
TYPE = functools.partial(custom_type, engine=None, controller=None)


def field(method, **kwargs):
    """Default Value For Type(s)"""
    default_value = None
    if callable(method):
        default_value = dc.field(default_factory=method, **kwargs)
    else:
        default_value = dc.field(default=method, **kwargs)
    return default_value


def load(models: list):
    """Init All Models"""
    types = {}
    for active in models:
        """
        is_component = is_model(active.cls)
        if is_component:
            if hasattr(active.cls, "__database__"):
                if callable(active.cls.__database__):
                    active.cls.__database__()
                types[active.key] = active.cls
        """
        app_name = get_module_name(active)
        module_uri = f"{app_name}.{active.__name__.lower()}"
        is_component = is_model(active)
        if is_component:
            if hasattr(active, "__database__"):
                if callable(active.__database__):
                    active.__database__()
                types[module_uri] = active
    return types


class Date:
    """Datetime for Testing"""

    @staticmethod
    def date():
        """Date"""
        return datetime.date.today()

    @staticmethod
    def datetime():
        """DateTime"""
        return datetime.datetime.now()

    @staticmethod
    def time():
        """Time"""
        return datetime.datetime.now().time()


class Controller:
    """Create Full-Controller"""

    def __init__(self, sql: str = None, mongo: str = None):
        if mongo:
            engine = "mongo"
            self.url = mongo
        elif sql:
            engine = "sql"
            self.url = sql
        self.engine = engine
        self.manager = create_database(sql=sql, mongo=mongo)

    @property
    def base(self):
        """Manager"""
        return self.manager.base

    @property
    def database(self):
        """Manager"""
        return self.manager.database

    def manage(self, objects=None):
        """Manager"""
        return self.manager.controller(objects=objects)

    @property
    def model(self):
        """SQL or Mongo"""
        return functools.partial(
            custom_type, engine=self.engine, controller=self.manager
        )



"""
    Mongo Manager
"""

# import functools
import math

from .utils import Objects  # clean_form, clean_update_form,
from .utils import Decode, Response, fixed_id_column, pagination


class MongoCrud:
    """
    Mongo Manager (CORE)
    """

    def __init__(self, client=None):
        self.collection = client

    async def create(self, form: dict | list) -> Response:
        """CREATE/CREATE-MANY"""
        if isinstance(form, dict):
            return await self.create_one(form)
        return await self.create_many(form)

    async def create_one(self, form: dict) -> Response:
        """CREATE"""
        collection = self.collection
        try:
            result = await collection.insert_one(form)
            instance = await collection.find_one({"_id": {"$eq": result.inserted_id}})
            result = Response(data=Objects.mongo(instance))
        except Exception as error:
            result = Response(error=True, error_message=str(error))
        return result

    async def create_many(self, forms: list) -> Response:
        """CREATE-MANY"""
        collection = self.collection
        try:
            result = await collection.insert_many(forms)
            search = {"_id": {"$in": result.inserted_ids}}
            cursor = collection.find(search)
            items = [i async for i in cursor]
            result = Response(data=Objects.mongo(items), count=len(items))
        except Exception as error:
            result = Response(error=True, error_message=str(error))
        return result

    async def update(self, search: dict = None, form: dict = None) -> Response:
        """UPDATE"""
        collection = self.collection
        try:
            instance = await collection.update_many(search, {"$set": form})
            count = instance.modified_count
            result = Response(count=count)
        except Exception as error:
            result = Response(error=True, error_message=str(error))
        return result

    async def delete(self, search: dict) -> Response:
        """DELETE"""
        collection = self.collection
        try:
            instance = await collection.delete_many(search)
            count = instance.deleted_count
            result = Response(count=count)
        except Exception as error:
            result = Response(error=True, error_message=str(error))
        return result

    async def find_one(self, search: dict) -> Response:
        """DETAIL"""
        collection = self.collection
        try:
            instance = await collection.find_one(search)
            result = Objects.mongo(instance)
        except Exception:
            result = None
        return result

    async def find(
        self,
        search: dict = None,
        page: int = 1,
        limit: int = 100,
        sort_by: str = "-id",
    ) -> Response:
        """FIND"""
        collection = self.collection
        sort_desc = 1
        # Check Sort By
        if sort_by.startswith("-"):
            sort_by = sort_by[1:]
            sort_desc = -1
        if sort_by == "id":
            sort_by = "_id"
        try:
            _page = pagination(page=page, limit=limit)
            # Add Sort By
            cursor = collection.find(search).sort(sort_by, sort_desc)
            # Offset & Limit
            cursor.skip(_page.offset).limit(_page.limit)
            items = [i async for i in cursor]
            count = await collection.count_documents(search)
            pages = int(math.ceil(count / limit))
            result = Response(data=Objects.mongo(items), count=count, pages=pages)
        except Exception as error:
            result = Response(error=True, error_message=str(error))
        return result

    async def all(self):
        """ALL-Rows"""
        collection = self.collection
        cursor = collection.find()
        items = [i async for i in cursor]
        return Response(data=Objects.mongo(items), count=len(items), pages=1)


class Mongo:
    """
    Mongo Manager (CRUD)
    """

    def __init__(self, objects=None, database=None, keys=None):
        self.collection = database
        self.crud = MongoCrud(objects)
        self.columns = keys

    @staticmethod
    def id_decode(unique_id):
        """ID-DECODER"""
        return Decode.mongo(unique_id)

    async def create(self, form: dict):
        """Create Single-Row."""
        return await self.crud.create(form)

    async def update(self, unique_ids: list[str], form: dict):
        """Update Multiple/Single-Row(s)"""
        if not isinstance(unique_ids, list):
            unique_ids = [unique_ids]
        _ids = [Decode.mongo(i) for i in unique_ids]
        search = {"_id": {"$in": _ids}}
        results = await self.crud.update(search, form)
        if results.count == 1 and not results.error:
            results.data = await self.detail(unique_ids[0])
        return results

    async def delete(self, unique_ids: list[str], all: bool = False):
        """Delete Multiple/Single-Row(s)"""
        if not all:
            if not isinstance(unique_ids, list):
                unique_ids = [unique_ids]
            _id = [Decode.mongo(i) for i in unique_ids]
            search = {"_id": {"$in": _id}}
        else:
            search = {}
        return await self.crud.delete(search)

    async def detail(self, unique_id: str):
        """Get Single-Row from Database Collection by ID"""
        _id = Decode.mongo(unique_id)
        search = {"_id": {"$eq": _id}}
        return await self.crud.find_one(search)

    async def get_by(self, **kwargs):
        """Get Single-Row by <Keyword-Arguments>"""
        kwargs = fixed_id_column(kwargs)
        search = {key: {"$eq": val} for key, val in kwargs.items()}
        return await self.crud.find_one(search)

    async def find_one(self, search: dict = None):
        """Get Single-Row from Database Collection"""
        return await self.crud.find_one(search)

    async def all(self):
        """Get All-Rows from Database"""
        return await self.crud.all()

    async def find(
        self, search: dict = None, page: int = 1, limit: int = 100, sort_by: str = "-id"
    ):
        """Get Multiple-Rows from Database Collection"""
        sort_by = fixed_id_column(sort_by)
        return await self.crud.find(
            search=search, page=page, limit=limit, sort_by=sort_by
        )

    async def filter_by(
        self, search: dict = None, page: int = 1, limit: int = 100, sort_by: str = "-id"
    ):
        """Get Multiple-Rows from Database Collection by <Keyword-Arguments>"""
        search = fixed_id_column(search)
        query = {key: {"$eq": val} for key, val in search.items()}
        return await self.find(search=query, page=page, limit=limit, sort_by=sort_by)

    async def search(
        self,
        columns: list | None = None,
        value: str | None = None,
        page: int = 1,
        limit: int = 100,
        sort_by: str = "-id",
    ):
        """Get Multiple-Rows from Database Collection by <Searching-Columns>"""
        search = [{key: {"$regex": value}} for key in columns]
        return await self.find({"$or": search}, page=page, limit=limit, sort_by=sort_by)



"""
    * Database Specific (Decode)
"""

try:
    from bson.objectid import ObjectId
except ImportError:
    ObjectId = None

from .ids import ID


def sql_id_decode(unique_id) -> int | None:
    """Decoder for SQL"""
    try:
        return_value = int(ID.decode(unique_id))
    except Exception:
        return_value = None
    return return_value


def mongo_id_decode(unique_id) -> str | None:
    """Decoder for Mongo"""
    try:
        return_value = ObjectId(ID.decode(unique_id))
    except Exception:
        return_value = None
    return return_value


class Decode:
    """Convert ID to <SQL or Mongo> ID's Type."""

    @staticmethod
    def mongo(graphql_id):
        """ID Decode to Mongo"""
        return mongo_id_decode(graphql_id)

    @staticmethod
    def sql(graphql_id):
        """ID Decode to SQL"""
        return sql_id_decode(graphql_id)



# Copyright 2021-present MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helpers for the 'hello' and legacy hello commands."""

import copy
import datetime
import itertools
from typing import Any, Generic, List, Mapping, Optional, Set, Tuple

from bson.objectid import ObjectId
from pymongo import common
from pymongo.server_type import SERVER_TYPE
from pymongo.typings import _DocumentType


class HelloCompat:
    CMD = "hello"
    LEGACY_CMD = "ismaster"
    PRIMARY = "isWritablePrimary"
    LEGACY_PRIMARY = "ismaster"
    LEGACY_ERROR = "not master"


def _get_server_type(doc):
    """Determine the server type from a hello response."""
    if not doc.get("ok"):
        return SERVER_TYPE.Unknown

    if doc.get("serviceId"):
        return SERVER_TYPE.LoadBalancer
    elif doc.get("isreplicaset"):
        return SERVER_TYPE.RSGhost
    elif doc.get("setName"):
        if doc.get("hidden"):
            return SERVER_TYPE.RSOther
        elif doc.get(HelloCompat.PRIMARY):
            return SERVER_TYPE.RSPrimary
        elif doc.get(HelloCompat.LEGACY_PRIMARY):
            return SERVER_TYPE.RSPrimary
        elif doc.get("secondary"):
            return SERVER_TYPE.RSSecondary
        elif doc.get("arbiterOnly"):
            return SERVER_TYPE.RSArbiter
        else:
            return SERVER_TYPE.RSOther
    elif doc.get("msg") == "isdbgrid":
        return SERVER_TYPE.Mongos
    else:
        return SERVER_TYPE.Standalone


class Hello(Generic[_DocumentType]):
    """Parse a hello response from the server.

    .. versionadded:: 3.12
    """

    __slots__ = ("_doc", "_server_type", "_is_writable", "_is_readable", "_awaitable")

    def __init__(self, doc: _DocumentType, awaitable: bool = False) -> None:
        self._server_type = _get_server_type(doc)
        self._doc: _DocumentType = doc
        self._is_writable = self._server_type in (
            SERVER_TYPE.RSPrimary,
            SERVER_TYPE.Standalone,
            SERVER_TYPE.Mongos,
            SERVER_TYPE.LoadBalancer,
        )

        self._is_readable = self.server_type == SERVER_TYPE.RSSecondary or self._is_writable
        self._awaitable = awaitable

    @property
    def document(self) -> _DocumentType:
        """The complete hello command response document.

        .. versionadded:: 3.4
        """
        return copy.copy(self._doc)

    @property
    def server_type(self) -> int:
        return self._server_type

    @property
    def all_hosts(self) -> Set[Tuple[str, int]]:
        """List of hosts, passives, and arbiters known to this server."""
        return set(
            map(
                common.clean_node,
                itertools.chain(
                    self._doc.get("hosts", []),
                    self._doc.get("passives", []),
                    self._doc.get("arbiters", []),
                ),
            )
        )

    @property
    def tags(self) -> Mapping[str, Any]:
        """Replica set member tags or empty dict."""
        return self._doc.get("tags", {})

    @property
    def primary(self) -> Optional[Tuple[str, int]]:
        """This server's opinion about who the primary is, or None."""
        if self._doc.get("primary"):
            return common.partition_node(self._doc["primary"])
        else:
            return None

    @property
    def replica_set_name(self) -> Optional[str]:
        """Replica set name or None."""
        return self._doc.get("setName")

    @property
    def max_bson_size(self) -> int:
        return self._doc.get("maxBsonObjectSize", common.MAX_BSON_SIZE)

    @property
    def max_message_size(self) -> int:
        return self._doc.get("maxMessageSizeBytes", 2 * self.max_bson_size)

    @property
    def max_write_batch_size(self) -> int:
        return self._doc.get("maxWriteBatchSize", common.MAX_WRITE_BATCH_SIZE)

    @property
    def min_wire_version(self) -> int:
        return self._doc.get("minWireVersion", common.MIN_WIRE_VERSION)

    @property
    def max_wire_version(self) -> int:
        return self._doc.get("maxWireVersion", common.MAX_WIRE_VERSION)

    @property
    def set_version(self) -> Optional[int]:
        return self._doc.get("setVersion")

    @property
    def election_id(self) -> Optional[ObjectId]:
        return self._doc.get("electionId")

    @property
    def cluster_time(self) -> Optional[Mapping[str, Any]]:
        return self._doc.get("$clusterTime")

    @property
    def logical_session_timeout_minutes(self) -> Optional[int]:
        return self._doc.get("logicalSessionTimeoutMinutes")

    @property
    def is_writable(self) -> bool:
        return self._is_writable

    @property
    def is_readable(self) -> bool:
        return self._is_readable

    @property
    def me(self) -> Optional[Tuple[str, int]]:
        me = self._doc.get("me")
        if me:
            return common.clean_node(me)
        return None

    @property
    def last_write_date(self) -> Optional[datetime.datetime]:
        return self._doc.get("lastWrite", {}).get("lastWriteDate")

    @property
    def compressors(self) -> Optional[List[str]]:
        return self._doc.get("compression")

    @property
    def sasl_supported_mechs(self) -> List[str]:
        """Supported authentication mechanisms for the current user.

        For example::

            >>> hello.sasl_supported_mechs
            ["SCRAM-SHA-1", "SCRAM-SHA-256"]

        """
        return self._doc.get("saslSupportedMechs", [])

    @property
    def speculative_authenticate(self) -> Optional[Mapping[str, Any]]:
        """The speculativeAuthenticate field."""
        return self._doc.get("speculativeAuthenticate")

    @property
    def topology_version(self) -> Optional[Mapping[str, Any]]:
        return self._doc.get("topologyVersion")

    @property
    def awaitable(self) -> bool:
        return self._awaitable

    @property
    def service_id(self) -> Optional[ObjectId]:
        return self._doc.get("serviceId")

    @property
    def hello_ok(self) -> bool:
        return self._doc.get("helloOk", False)



# -*- coding: utf-8 -*-
"""
    API - GraphQL
"""

# Fastberry
import fastberry as fb

# Type(s) Tools
from .. import forms, manager, types

# Create your API (GraphQL) here.
@fb.gql
class Todo:
    """Todo Api"""

    class Meta:
        """Meta-Data"""

        app = False
        model = None

    class Query:
        """Query"""

        async def search(status: str | None = None) -> fb.edges(types.Task):
            """Search for task(s) by status"""
            return await manager.Task.search(status)

        async def detail(item: fb.ID) -> fb.query(types.Task):
            """Get task by ID"""
            return await manager.Task.detail(item)

    class Mutation:
        """Mutation"""

        async def create(form: forms.Task) -> fb.editor(types.Task):
            """Create a new task"""
            return await manager.Task.create(form.input)

        async def update(item: fb.ID, status: str) -> fb.mutation(types.Task):
            """Change task's status"""
            return await manager.Task.update(item, status)

        async def delete(item: fb.ID) -> bool:
            """Delete task by ID"""
            return await manager.Task.delete(item)



"""
    * SQLAlchemy + Databases — Controller
"""
# import functools
import math

from .utils import Objects  # clean_form, clean_update_form,
from .utils import Decode, Response, fixed_id_column
from .utils.sql_where import Filters as SQLFilters
from functools import partial

try:
    from sqlalchemy.sql.elements import BinaryExpression
except ImportError:
    BinaryExpression = None


class SQL:
    """SQlAlchemy & Databases (Manager)

    -----------------------------------------------------------------------------------------------
    # Init
    -----------------------------------------------------------------------------------------------
    sql = SQLBase(database_url, custom_type)

    -----------------------------------------------------------------------------------------------
    # Form
    -----------------------------------------------------------------------------------------------
    - Clean-Inputs      : sql.form(dict)
    - Clean-Empty       : sql.form_update(dict)

    -----------------------------------------------------------------------------------------------
    # Read (Examples)
    -----------------------------------------------------------------------------------------------
    - Get-By(Single-Row): await sql.get_by(id=1, name="spongebob")
    - Custom-Find-One   : await sql.find_one(sql.Q.where("id", "eq", 1))
    - Filter-By         : await sql.filter_by(id=1, name="spongebob", page=1, etc...)
    - Search-Columns    : await sql.search(["name", "title"], "bob", page=1, etc...)
    - Custom-Querying   : await sql.find(
                                    sql.Q.where("id", "in", [1, 2, 3]),
                                    page=1, limit=100, sort_by='-id'
                                )

    -----------------------------------------------------------------------------------------------
    # C.U.D
    -----------------------------------------------------------------------------------------------
    - Create            : await sql.create(dict)
    - Update            : await sql.update(items: list[IDs], values: dict)
    - Delete            : await sql.delete(items: list[IDs])
    """

    def __init__(self, objects=None, database=None):
        """Start Manager"""
        self.database = database
        self.table = objects
        self.Q = SQLFilters(objects)
        self.where = self.Q.where
        self.columns = self.table.columns.keys()
        self.c = self.table.c
        self.to_obj = partial(Objects.sql, self.columns)

    @staticmethod
    def id_decode(unique_id):
        """ID-DECODER"""
        return Decode.sql(unique_id)

    async def create(
        self, form: dict | list = None, return_items: bool = True
    ) -> Response:
        """CREATE/CREATE-MANY"""
        if isinstance(form, dict):
            return await self.create_one(form)
        return await self.create_many(form, return_items)

    async def _create_one_row(self, form: dict) -> Response:
        """(Base) Create Single-Row."""
        return_value = Response()
        try:
            sql_query = self.table.insert(form)
            return_value.data = await self.database.execute(sql_query)
        except Exception as error:
            return_value.error = True
            return_value.error_message = str(error)
        return return_value

    async def create_one(self, form: dict):
        """Create Single-Row."""
        return_value = await self._create_one_row(form)
        if return_value.data and not return_value.error:
            return_value.data = await self.get_by(_id=return_value.data)
        return return_value

    async def _create_many_rows(self, form: list):
        """(Base) Create Multiple-Row."""
        return_value = Response()
        try:
            sql_query = self.table.insert()
            unique_ids = await self.database.execute_many(sql_query, values=form)
            return_value.data = unique_ids
            return_value.count = len(form)
        except Exception as error:
            return_value.error = True
            return_value.error_message = str(error)
        return return_value

    async def create_many(self, forms: list, return_items: bool):
        """Create Single-Row."""
        if return_items:
            all_ids = []
            for form in forms:
                return_value = await self._create_one_row(form)
                if return_value.data:
                    all_ids.append(return_value.data)
            sql_ids_in = self.Q.where("_id", "in", all_ids)
            items = await self.database.fetch_all(self.Q.select(sql_ids_in))
            return Response(data=self.to_obj(items), count=len(items))
        return await self._create_many_rows(forms)

    async def update(self, unique_ids: list[str], form: dict):
        """Update Multiple/Single-Row(s)"""
        return_value = Response()
        # Get Ids
        if not isinstance(unique_ids, list):
            unique_ids = [unique_ids]
        all_ids = [Decode.sql(i) for i in unique_ids]
        sql_ids_in = self.Q.where("_id", "in", all_ids)
        try:
            return_value.count = await self.database.execute(
                self.table.update().where(sql_ids_in).values(**form)
            )
        except Exception as error:
            return_value.error = True
            return_value.error_message = str(error)
        # Get Details
        if len(all_ids) == 1 and return_value.count == 1:
            return_value.data = await self.get_by(_id=all_ids[0])
        return return_value

    async def delete(self, unique_ids: list[str], all: bool = False):
        """Delete Multiple/Single-Row(s)"""
        return_value = Response()
        # Get Ids
        if not isinstance(unique_ids, list):
            unique_ids = [unique_ids]
        if not all:
            all_ids = [Decode.sql(i) for i in unique_ids]
            sql_ids_in = self.Q.where("_id", "in", all_ids)
            selector = self.table.delete().where(sql_ids_in)
        else:
            selector = self.table.delete()
        try:
            return_value.count = await self.database.execute(selector)
        except Exception as error:
            return_value.error = True
            return_value.error_message = str(error)
        return return_value

    async def detail(self, ID):
        """Get Single-Row from Database Table by ID"""
        query = self.Q.filter_by(_id=Decode.sql(ID))
        item = await self.database.fetch_one(self.Q.select(query))
        return self.to_obj(item)

    async def get_by(self, **kwargs):
        """Get Single-Row from Database Table by <Keyword-Arguments>"""
        kwargs = fixed_id_column(kwargs)
        query = self.Q.filter_by(**kwargs)
        item = await self.database.fetch_one(self.Q.select(query))
        return self.to_obj(item)

    async def find_one(self, query):
        """Get Single-Row from Database Table by <SQLAlchemy-BinaryExpression>"""
        item = await self.database.fetch_one(self.Q.select(query))
        return self.to_obj(item)

    async def all(
        self,
    ):
        """Get All-Rows from Database Table"""
        items = await self.database.fetch_all(self.Q.select())
        return Response(data=self.to_obj(items), count=len(items), pages=1)

    async def find(
        self,
        search: BinaryExpression = None,
        page: int | None = None,
        limit: int | None = None,
        sort_by: str | None = None,
    ):
        """Get Multiple-Rows from Database Table by <SQLAlchemy-BinaryExpression>"""
        sort_by = fixed_id_column(sort_by)
        try:
            query = self.Q.find(search, page=page, limit=limit, sort_by=sort_by)
            items = await self.database.fetch_all(query.query)
            count = await self.database.fetch_val(query.count)
            _limit = limit or 1
            pages = int(math.ceil(count / _limit))
            return Response(data=self.to_obj(items), count=count, pages=pages)
        except Exception as e:
            print(e)
            return Response(data=[], count=0, pages=0)

    async def filter_by(
        self,
        search: dict | None = None,
        page: int | None = None,
        limit: int | None = None,
        sort_by: str | None = None,
    ):
        """Get Multiple-Rows from Database Table by <Keyword-Arguments>"""
        search = fixed_id_column(search)
        query = self.Q.filter_by(**search) if search else None
        items = await self.find(query, page=page, limit=limit, sort_by=sort_by)
        return items

    async def search(
        self,
        columns: list | None = None,
        value: str | None = None,
        page: int | None = None,
        limit: int | None = None,
        sort_by: str | None = None,
    ):
        """Get Multiple-Rows from Database Table by <Searching-Columns>"""
        query = self.Q.search(columns, value) if value else None
        items = await self.find(query, page=page, limit=limit, sort_by=sort_by)
        return items



# Copyright (C) Dnspython Contributors, see LICENSE for text of ISC license

import struct

import dns.immutable


@dns.immutable.immutable
class L32(dns.rdata.Rdata):

    """L32 record"""

    # see: rfc6742.txt

    __slots__ = ['preference', 'locator32']

    def __init__(self, rdclass, rdtype, preference, locator32):
        super().__init__(rdclass, rdtype)
        self.preference = self._as_uint16(preference)
        self.locator32 = self._as_ipv4_address(locator32)

    def to_text(self, origin=None, relativize=True, **kw):
        return f'{self.preference} {self.locator32}'

    @classmethod
    def from_text(cls, rdclass, rdtype, tok, origin=None, relativize=True,
                  relativize_to=None):
        preference = tok.get_uint16()
        nodeid = tok.get_identifier()
        return cls(rdclass, rdtype, preference, nodeid)

    def _to_wire(self, file, compress=None, origin=None, canonicalize=False):
        file.write(struct.pack('!H', self.preference))
        file.write(dns.ipv4.inet_aton(self.locator32))

    @classmethod
    def from_wire_parser(cls, rdclass, rdtype, parser, origin=None):
        preference = parser.get_uint16()
        locator32 = parser.get_remaining()
        return cls(rdclass, rdtype, preference, locator32)



# Copyright 2015-present MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Internal network layer helper methods."""

import datetime
import errno
import socket
import struct
import time

from bson import _decode_all_selective
from pymongo import _csot, helpers, message, ssl_support
from pymongo.common import MAX_MESSAGE_SIZE
from pymongo.compression_support import _NO_COMPRESSION, decompress
from pymongo.errors import (
    NotPrimaryError,
    OperationFailure,
    ProtocolError,
    _OperationCancelled,
)
from pymongo.message import _UNPACK_REPLY, _OpMsg
from pymongo.monitoring import _is_speculative_authenticate
from pymongo.socket_checker import _errno_from_exception

_UNPACK_HEADER = struct.Struct("<iiii").unpack


def command(
    sock_info,
    dbname,
    spec,
    is_mongos,
    read_preference,
    codec_options,
    session,
    client,
    check=True,
    allowable_errors=None,
    address=None,
    listeners=None,
    max_bson_size=None,
    read_concern=None,
    parse_write_concern_error=False,
    collation=None,
    compression_ctx=None,
    use_op_msg=False,
    unacknowledged=False,
    user_fields=None,
    exhaust_allowed=False,
    write_concern=None,
):
    """Execute a command over the socket, or raise socket.error.

    :Parameters:
      - `sock`: a raw socket instance
      - `dbname`: name of the database on which to run the command
      - `spec`: a command document as an ordered dict type, eg SON.
      - `is_mongos`: are we connected to a mongos?
      - `read_preference`: a read preference
      - `codec_options`: a CodecOptions instance
      - `session`: optional ClientSession instance.
      - `client`: optional MongoClient instance for updating $clusterTime.
      - `check`: raise OperationFailure if there are errors
      - `allowable_errors`: errors to ignore if `check` is True
      - `address`: the (host, port) of `sock`
      - `listeners`: An instance of :class:`~pymongo.monitoring.EventListeners`
      - `max_bson_size`: The maximum encoded bson size for this server
      - `read_concern`: The read concern for this command.
      - `parse_write_concern_error`: Whether to parse the ``writeConcernError``
        field in the command response.
      - `collation`: The collation for this command.
      - `compression_ctx`: optional compression Context.
      - `use_op_msg`: True if we should use OP_MSG.
      - `unacknowledged`: True if this is an unacknowledged command.
      - `user_fields` (optional): Response fields that should be decoded
        using the TypeDecoders from codec_options, passed to
        bson._decode_all_selective.
      - `exhaust_allowed`: True if we should enable OP_MSG exhaustAllowed.
    """
    name = next(iter(spec))
    ns = dbname + ".$cmd"
    speculative_hello = False

    # Publish the original command document, perhaps with lsid and $clusterTime.
    orig = spec
    if is_mongos and not use_op_msg:
        spec = message._maybe_add_read_preference(spec, read_preference)
    if read_concern and not (session and session.in_transaction):
        if read_concern.level:
            spec["readConcern"] = read_concern.document
        if session:
            session._update_read_concern(spec, sock_info)
    if collation is not None:
        spec["collation"] = collation

    publish = listeners is not None and listeners.enabled_for_commands
    if publish:
        start = datetime.datetime.now()
        speculative_hello = _is_speculative_authenticate(name, spec)

    if compression_ctx and name.lower() in _NO_COMPRESSION:
        compression_ctx = None

    if client and client._encrypter and not client._encrypter._bypass_auto_encryption:
        spec = orig = client._encrypter.encrypt(dbname, spec, codec_options)

    # Support CSOT
    if client:
        sock_info.apply_timeout(client, spec)
    _csot.apply_write_concern(spec, write_concern)

    if use_op_msg:
        flags = _OpMsg.MORE_TO_COME if unacknowledged else 0
        flags |= _OpMsg.EXHAUST_ALLOWED if exhaust_allowed else 0
        request_id, msg, size, max_doc_size = message._op_msg(
            flags, spec, dbname, read_preference, codec_options, ctx=compression_ctx
        )
        # If this is an unacknowledged write then make sure the encoded doc(s)
        # are small enough, otherwise rely on the server to return an error.
        if unacknowledged and max_bson_size is not None and max_doc_size > max_bson_size:
            message._raise_document_too_large(name, size, max_bson_size)
    else:
        request_id, msg, size = message._query(
            0, ns, 0, -1, spec, None, codec_options, compression_ctx
        )

    if max_bson_size is not None and size > max_bson_size + message._COMMAND_OVERHEAD:
        message._raise_document_too_large(name, size, max_bson_size + message._COMMAND_OVERHEAD)

    if publish:
        encoding_duration = datetime.datetime.now() - start
        listeners.publish_command_start(
            orig, dbname, request_id, address, service_id=sock_info.service_id
        )
        start = datetime.datetime.now()

    try:
        sock_info.sock.sendall(msg)
        if use_op_msg and unacknowledged:
            # Unacknowledged, fake a successful command response.
            reply = None
            response_doc = {"ok": 1}
        else:
            reply = receive_message(sock_info, request_id)
            sock_info.more_to_come = reply.more_to_come
            unpacked_docs = reply.unpack_response(
                codec_options=codec_options, user_fields=user_fields
            )

            response_doc = unpacked_docs[0]
            if client:
                client._process_response(response_doc, session)
            if check:
                helpers._check_command_response(
                    response_doc,
                    sock_info.max_wire_version,
                    allowable_errors,
                    parse_write_concern_error=parse_write_concern_error,
                )
    except Exception as exc:
        if publish:
            duration = (datetime.datetime.now() - start) + encoding_duration
            if isinstance(exc, (NotPrimaryError, OperationFailure)):
                failure = exc.details
            else:
                failure = message._convert_exception(exc)
            listeners.publish_command_failure(
                duration, failure, name, request_id, address, service_id=sock_info.service_id
            )
        raise
    if publish:
        duration = (datetime.datetime.now() - start) + encoding_duration
        listeners.publish_command_success(
            duration,
            response_doc,
            name,
            request_id,
            address,
            service_id=sock_info.service_id,
            speculative_hello=speculative_hello,
        )

    if client and client._encrypter and reply:
        decrypted = client._encrypter.decrypt(reply.raw_command_response())
        response_doc = _decode_all_selective(decrypted, codec_options, user_fields)[0]

    return response_doc


_UNPACK_COMPRESSION_HEADER = struct.Struct("<iiB").unpack


def receive_message(sock_info, request_id, max_message_size=MAX_MESSAGE_SIZE):
    """Receive a raw BSON message or raise socket.error."""
    if _csot.get_timeout():
        deadline = _csot.get_deadline()
    else:
        timeout = sock_info.sock.gettimeout()
        if timeout:
            deadline = time.monotonic() + timeout
        else:
            deadline = None
    # Ignore the response's request id.
    length, _, response_to, op_code = _UNPACK_HEADER(
        _receive_data_on_socket(sock_info, 16, deadline)
    )
    # No request_id for exhaust cursor "getMore".
    if request_id is not None:
        if request_id != response_to:
            raise ProtocolError("Got response id %r but expected %r" % (response_to, request_id))
    if length <= 16:
        raise ProtocolError(
            "Message length (%r) not longer than standard message header size (16)" % (length,)
        )
    if length > max_message_size:
        raise ProtocolError(
            "Message length (%r) is larger than server max "
            "message size (%r)" % (length, max_message_size)
        )
    if op_code == 2012:
        op_code, _, compressor_id = _UNPACK_COMPRESSION_HEADER(
            _receive_data_on_socket(sock_info, 9, deadline)
        )
        data = decompress(_receive_data_on_socket(sock_info, length - 25, deadline), compressor_id)
    else:
        data = _receive_data_on_socket(sock_info, length - 16, deadline)

    try:
        unpack_reply = _UNPACK_REPLY[op_code]
    except KeyError:
        raise ProtocolError("Got opcode %r but expected %r" % (op_code, _UNPACK_REPLY.keys()))
    return unpack_reply(data)


_POLL_TIMEOUT = 0.5


def wait_for_read(sock_info, deadline):
    """Block until at least one byte is read, or a timeout, or a cancel."""
    context = sock_info.cancel_context
    # Only Monitor connections can be cancelled.
    if context:
        sock = sock_info.sock
        timed_out = False
        while True:
            # SSLSocket can have buffered data which won't be caught by select.
            if hasattr(sock, "pending") and sock.pending() > 0:
                readable = True
            else:
                # Wait up to 500ms for the socket to become readable and then
                # check for cancellation.
                if deadline:
                    remaining = deadline - time.monotonic()
                    # When the timeout has expired perform one final check to
                    # see if the socket is readable. This helps avoid spurious
                    # timeouts on AWS Lambda and other FaaS environments.
                    if remaining <= 0:
                        timed_out = True
                    timeout = max(min(remaining, _POLL_TIMEOUT), 0)
                else:
                    timeout = _POLL_TIMEOUT
                readable = sock_info.socket_checker.select(sock, read=True, timeout=timeout)
            if context.cancelled:
                raise _OperationCancelled("hello cancelled")
            if readable:
                return
            if timed_out:
                raise socket.timeout("timed out")


# Errors raised by sockets (and TLS sockets) when in non-blocking mode.
BLOCKING_IO_ERRORS = (BlockingIOError,) + ssl_support.BLOCKING_IO_ERRORS


def _receive_data_on_socket(sock_info, length, deadline):
    buf = bytearray(length)
    mv = memoryview(buf)
    bytes_read = 0
    while bytes_read < length:
        try:
            wait_for_read(sock_info, deadline)
            # CSOT: Update timeout. When the timeout has expired perform one
            # final non-blocking recv. This helps avoid spurious timeouts when
            # the response is actually already buffered on the client.
            if _csot.get_timeout():
                sock_info.set_socket_timeout(max(deadline - time.monotonic(), 0))
            chunk_length = sock_info.sock.recv_into(mv[bytes_read:])
        except BLOCKING_IO_ERRORS:
            raise socket.timeout("timed out")
        except (IOError, OSError) as exc:  # noqa: B014
            if _errno_from_exception(exc) == errno.EINTR:
                continue
            raise
        if chunk_length == 0:
            raise OSError("connection closed")

        bytes_read += chunk_length

    return mv



"""
    [Front-end Builder]
"""

import json
from pathlib import Path
from types import SimpleNamespace

INTROSPECT_OBJECTS = """query IntrospectionQuery{__schema{queryType{name}mutationType{name}subscriptionType{name}types{...FullType}directives{name description locations args{...InputValue}}}}fragment FullType on __Type{kind name description fields(includeDeprecated:true){name description args{...InputValue}type{...TypeRef}isDeprecated deprecationReason}inputFields{...InputValue}interfaces{...TypeRef}enumValues(includeDeprecated:true){name description isDeprecated deprecationReason}possibleTypes{...TypeRef}}fragment InputValue on __InputValue{name description type{...TypeRef}defaultValue}fragment TypeRef on __Type{kind name ofType{kind name ofType{kind name ofType{kind name ofType{kind name ofType{kind name ofType{kind name}}}}}}}"""

# schema.execute_sync(INTROSPECT_OBJECTS, variable_values=None)
POSSIBLE_TYPES = {
    "schema": ["queryType", "mutationType", "subscriptionType", "types", "directives"],
    "types": ["OBJECT", "SCALAR", "ENUM", "INPUT_OBJECT", "UNION"],
    "objects": [
        "kind",
        "name",
        "description",
        "fields",
        "inputFields",
        "interfaces",
        "enumValues",
        "possibleTypes",
    ],
}


def graphql_introspect_objects(schema):
    """Introspection"""
    results = schema.execute_sync(INTROSPECT_OBJECTS, variable_values=None)
    return results.data.get("__schema", {})


def get_types(schema):
    """GraphQL Types"""
    out_schema = {}
    for item in schema.get("types", []):
        kind = item.get("kind")
        if out_schema.get(kind):
            out_schema[kind].append(item)
        else:
            out_schema[kind] = []
            out_schema[kind].append(item)
    return out_schema


def transform_field(field):
    """GraphQL Fields"""
    if not field.get("required"):
        field["required"] = False
    if not field.get("list"):
        field["list"] = False
    if not field.get("defaultValue"):
        field["defaultValue"] = None
    return field


def get_field_info(field):
    """GraphQL Field Info"""
    current = SimpleNamespace(**transform_field(field))
    if not current.ofType:
        return current
    current_dict = get_field_info(current.ofType)
    # Trying to Fix IT
    if current.kind == "NON_NULL":
        current_dict.required = True
    if current.kind == "LIST":
        current_dict.list = True
    return current_dict


def get_object_fields(fields):
    """GraphQL Object"""
    items = {}
    for field in fields:
        field = SimpleNamespace(**field)
        field_kind = field.type.get("kind")
        current_field = get_field_info(field.type)
        current_field = current_field.__dict__
        del current_field["ofType"]
        if "SCALAR" == current_field.get("kind") and field_kind == "NON_NULL":
            current_field["required"] = True
        if field_kind == "LIST":
            current_field["list"] = True
        items[field.name] = current_field
    return items


def get_input_config(all_args):
    """GraphQL Input Config"""
    field_list = []
    form_dict = {}
    for field in all_args:
        # Step 1
        current = get_field_info(field["type"])
        if field["defaultValue"] in ["null", "None", None]:
            default_value = "null"
        else:
            default_value = field["defaultValue"]
        # Step 2
        try:
            default_value = json.loads(default_value)
        except:
            default_value = None
        # Step 3
        current.defaultValue = default_value
        field_dict = current.__dict__
        del field_dict["ofType"]
        # Append
        form_dict[field["name"]] = field_dict["defaultValue"]
        field_list.append(
            {
                "name": field["name"],
                "type": field_dict,
            }
        )
    return field_list, form_dict


def get_op_args(operations):
    """GraphQL Operations"""
    items = {}
    for item in operations:
        field_list, form_dict = get_input_config(item.get("args"))
        items[item.get("name")] = {
            "args": field_list,
            "form": form_dict,
        }
    return items


def get_graphql_objects_base(graphql_objects):
    """GraphQL Starter"""
    all_graphql_objects = {}
    for item in graphql_objects:
        name = item.get("name")
        fields = item.get("fields")
        if not name.startswith("__") and name not in ["Query", "Mutation"]:
            current_config = get_object_fields(fields)
            for key, conf in current_config.items():
                del conf["defaultValue"]
            all_graphql_objects[name] = current_config
        elif name in ["Query", "Mutation"]:
            obj_types = get_object_fields(fields)
            obj_args = get_op_args(fields)
            all_graphql_objects[name] = {}
            for key, val in obj_types.items():
                all_graphql_objects[name][key] = {"type": val, **obj_args[key]}
    return all_graphql_objects


def get_graphql_objects(all_objects):
    """GraphQL Main Objects"""
    all_objs = get_graphql_objects_base(all_objects)
    all_query = all_objs.get("Query")
    all_mutation = all_objs.get("Mutation")
    if all_query:
        del all_objs["Query"]
    if all_mutation:
        del all_objs["Mutation"]
    return {
        "OBJECT": all_objs,
        "Query": all_query,
        "Mutation": all_mutation,
    }


def get_graphql_input_objects(graphql_input_objects):
    """GraphQL Input Objects"""
    dict_out = {}
    for item in graphql_input_objects:
        field_list, form_dict = get_input_config(item["inputFields"])
        # Dict Inject
        dict_out[item["name"]] = {
            "args": field_list,
            "form": form_dict,
        }
    return dict_out


def get_graphql_scalars(graphql_scalars):
    """GraphQL Scalar"""
    all_graphql_scalars = set()
    for item in graphql_scalars:
        all_graphql_scalars.add(item.get("name"))
    return list(all_graphql_scalars)


def get_graphql_unions(graphql_unions):
    """GraphQL Union"""
    all_graphql_unions = {}
    for item in graphql_unions:
        item = SimpleNamespace(**item)
        all_graphql_unions[item.name] = {}
        for gtype in item.possibleTypes:
            current = transform_field(
                {
                    "kind": gtype.get("kind"),
                    "name": gtype.get("name"),
                }
            )
            del current["defaultValue"]
            all_graphql_unions[item.name][gtype.get("name")] = current
    return all_graphql_unions


def get_graphql_enum(all_enum):
    """GraphQL Enum"""
    all_graphql_enum = {}
    for item in all_enum:
        name = item.get("name")
        if not name.startswith("__"):
            all_graphql_enum[name] = [val.get("name") for val in item.get("enumValues")]
    return all_graphql_enum


def transform_operations(schema, current):
    """GraphQL Transform Operations 1"""
    current_type = current["type"]
    current_form = current["form"]
    current_args = current["args"]
    for field in current_args:
        if field["type"]["kind"] == "INPUT_OBJECT":
            input_form = (
                schema["INPUT_OBJECT"].get(field["type"]["name"], {}).get("form")
            )
            current_form[field["name"]] = input_form
    return {"type": current_type, "args": current_args, "form": current_form}


def get_all_transform_operations(schema, op_type):
    """GraphQL Transform Operations 2"""
    query_ops = {}
    all_query = schema.get(op_type, {})
    for op in all_query.keys():
        current = all_query.get(op)
        query_ops[op] = transform_operations(schema, current)
    return query_ops


def parse_graphql_schema(schema, include_input_object=True):
    """GraphQL Parse Schema"""
    # Step 1
    schema_introspect = graphql_introspect_objects(schema)
    schema_types = get_types(schema_introspect)
    return_object = get_graphql_objects(schema_types["OBJECT"])
    # Step 2
    schema_dict = {
        **return_object,
        "UNION": get_graphql_unions(schema_types.get("UNION")),
        "SCALAR": get_graphql_scalars(schema_types.get("SCALAR")),
        "INPUT_OBJECT": get_graphql_input_objects(schema_types.get("INPUT_OBJECT")),
        "ENUM": get_graphql_enum(schema_types.get("ENUM")),
    }
    # Step 3
    schema_dict["Query"] = get_all_transform_operations(schema_dict, "Query")
    schema_dict["Mutation"] = get_all_transform_operations(schema_dict, "Mutation")
    # Step 4
    if not include_input_object:
        del schema_dict["INPUT_OBJECT"]
    # Schema Out
    return schema_dict


def write_file(path, data):
    """Write To File"""
    file_path = Path(path)
    parent_dir = file_path.parent
    if not parent_dir.exists():
        parent_dir.mkdir(parents=True)
    with open(file_path, "w", encoding="utf-8") as file:
        file.write(data)


def create_frontend_objects(all_objs):
    """Frontend Objects"""
    frontend_setup = {}
    for name, config in all_objs.items():
        config_js = {}
        for field_name, field_config in config.items():
            match field_config["kind"]:
                case "SCALAR":
                    config_js[field_name] = True
                case "OBJECT":
                    # "#" +
                    config_js[field_name] = field_config["name"]
        frontend_setup[name] = config_js

    return frontend_setup


def create_frontend_operations(schema, all_objs):
    """Frontend Operations"""
    query_frontend = {}
    for name, config in all_objs.items():
        return_value = None
        match config["type"]["kind"]:
            case "SCALAR":
                return_value = None
            case "OBJECT":
                return_value = [config["type"]["name"]]
            case "UNION":
                child_name = config["type"]["name"]
                return_value = list(schema["UNION"][child_name].keys())
        query_frontend[name] = {
            "model": return_value,
            "args": config["args"],
            "form": config["form"],
        }
    return query_frontend


class Frontend:
    """Frontend Builder"""

    def __init__(self, schema):
        self._schema = parse_graphql_schema(schema, True)

    @property
    def query(self):
        """All Queries"""
        items = self._schema.get("Query")
        if items:
            return create_frontend_operations(self._schema, items)
        return {}

    @property
    def mutation(self):
        """All Mutations"""
        items = self._schema.get("Mutation")
        if items:
            return create_frontend_operations(self._schema, items)
        return {}

    @property
    def models(self):
        """All Models"""
        items = self._schema.get("OBJECT")
        if items:
            schema = create_frontend_objects(items)
            return json.dumps(schema, indent=4)
        return {}

    @property
    def schema(self):
        """Schema"""
        return json.dumps(
            {
                "query": self.query,
                "mutation": self.mutation,
                "enum": self._schema.get("ENUM"),
                "scalar": self._schema.get("SCALAR"),
                "forms": self._schema.get("INPUT_OBJECT"),
            },
            indent=0,
        )



"""
    Class Tools
"""


class ClassPropertyDescriptor:
    """Used to generate the @classproperty"""

    def __init__(self, fget, fset=None):
        self.fget = fget
        self.fset = fset

    def __get__(self, obj, klass=None):
        if klass is None:
            klass = type(obj)
        return self.fget.__get__(obj, klass)()

    def __set__(self, obj, value):
        if not self.fset:
            raise AttributeError("can't set attribute")
        type_ = type(obj)
        return self.fset.__get__(obj, type_)(value)

    def setter(self, func):
        """Set Class-Property"""
        if not isinstance(func, (classmethod, staticmethod)):
            func = classmethod(func)
        self.fset = func
        return self


def classproperty(func):
    """Use to create a Class-Property"""

    if not isinstance(func, (classmethod, staticmethod)):
        func = classmethod(func)
    return ClassPropertyDescriptor(func)


"""
    Model Querying
"""


class Manager:
    """Class: Create a Manager"""

    model = None

    @classproperty
    def engine(cls):
        """Model's Engine"""
        return cls.model.__database__

    @classproperty
    def objects(cls):
        """Model's Controller"""
        return cls.model.objects

    @classmethod
    def form(cls, __remove__: list | None, **kwargs):
        """
        Remove all { Values } that are { None }
        Plus any { Field } in the { __remove__ } section.
        """
        active_filter = __remove__ or []
        instance = cls.model(**kwargs)
        form = {
            key: val
            for key, val in instance.__dict__.items()
            if val is not None and key not in active_filter
        }
        return form


def manager(cls):
    """Function: Create a Singleton"""
    custom_class = type(
        f"{cls.__name__}Manager",
        (cls, Manager),
        {},
    )
    return custom_class



""" Strawberry - GraphQL (Connection, PageInfo, Edges)
    Made Easy by wrapping functions together.

methods:
    Mutation: GraphQL Mutation.
    Query: GraphQL Query.
    Edges: GraphQL Edges.
    Response: Response GraphQL Edges.
    Error: Response Error([ErrorMessage]).
    ErrorMessage: Error Message.
"""

from dataclasses import field

import strawberry
from strawberry.scalars import JSON

from .edges import Connection, Edge, PageInfo


@strawberry.type
class ErrorMessage:
    """GraphQL Error Message"""

    # ['field', 'type', 'text']

    field: str | None = None
    type: str | None = None
    text: str | None = None


@strawberry.type
class Error:
    """GraphQL Error"""

    messages: list[ErrorMessage] = field(default_factory=list)
    meta: JSON = field(default_factory=dict)
    error: bool = True


@strawberry.type
class Deleted:
    """GraphQL Deleted"""

    count: int = 0


def Query(model):
    """GraphQL Query"""
    return model | None


def Mutation(model):
    """GraphQL Mutation"""
    return model | Error


def Editor(model):
    """GraphQL Editor aka: (Super-Mutation)"""
    return model | Error | Deleted


def Edges(model) -> Connection:
    """GraphQL Edges"""
    return Connection[model]


def Page(
    edges: list[Edge] = None, length: int = 0, pages: int = 0, extra: dict = None
) -> Connection:
    """Edges { Page } Response"""
    edges = edges or []
    extra = extra or {}
    items = [Edge(node=item, cursor=item.id) for item in edges]
    return Connection(
        page_info=PageInfo(
            length=length,
            pages=pages,
            extra=extra,
        ),
        edges=items,
    )
